## code to prepare `DATASET` dataset goes here

region_at_time_txtfls = function(filename){
  # function returning the mapping from one region name and code tag to the next according to the SSB region classification
  #Fylker: https://www.ssb.no/en/klass/klassifikasjoner/104/versjon/1158/koder
  #Kommuner: https://www.ssb.no/en/klass/klassifikasjoner/131

  #   filename = files[1]
  #readr::guess_encoding(filename)
  datastring = readLines(filename, n=-1L,
                         encoding =  dplyr::pull(readr::guess_encoding(filename)[1,1]),
                         warn = F)#nchars = 10^6)
  headings = unlist(stringr::str_split(datastring[1], "\t"))
  print(headings)
  datastring = datastring[-1]

  if((length(headings)%% 2) == 0 & length(datastring)>1) {
    # Then string should be arranged to pairs of "froms" and "tos"
    datastring = stringr::str_remove(datastring, pattern = "\t")
    convtable = matrix(data = datastring, ncol = 2, byrow = T)
    colnames(convtable) = headings
    headingsinv = unlist(lapply(X= str_split(headings, " "), FUN = function(X){paste0(X[2]," ",  X[1], " 1")}))

    colnamecandidates = stringr::str_sub(
      stringr::str_replace_all(
        string = lubridate::ymd(headingsinv),
        pattern = "-",
        replacement = ""),
      start=1, end = 6)

    convtable = dplyr::as_tibble(as.data.frame(convtable, stringsAsFactors=F))
    froms =  dplyr::as_tibble(
      str_split(
        string =  dplyr::pull(convtable[,1]), pattern = " - ", n=2, simplify = T))
    colnames(froms) = paste(c("reg_code", "reg_name"), rep(colnamecandidates[1], 2), sep = "_")

    tos = dplyr::as_tibble(
      stringr::str_split(
        string =  dplyr::pull(convtable[,2]), pattern = " - ", n=2, simplify = T))
    colnames(tos) = paste(c("reg_code", "reg_name"), rep(colnamecandidates[2], 2), sep = "_")

    fromstos = dplyr::bind_cols(froms, tos)
  } else if((length(headings)%% 2) == 1 & length(datastring)>1) {
    # Then it is the starting point, i.e, first array of region units

    convtable = matrix(data = datastring, ncol = 1, byrow = T)
    #colnames(convtable) = headings
    headingsinv =
      unlist(lapply(X= stringr::str_split(headings, " "),
                    FUN = function(X){ paste0(X[2]," ",  X[1], " 1")
                    }))
    colnamecandidates =
      stringr::str_sub(
        stringr::str_replace_all(
          string =  lubridate::ymd(headingsinv),
          pattern = "-",
          replacement = ""), start=1, end = 6)

    convtable = as.data.frame(convtable, stringsAsFactors=F )
    froms =  data.frame(
      stringr::str_split(
        string =  convtable[,1],
        pattern = " - ",
        n=2,
        simplify = T), stringsAsFactors = F )

    colnames(froms) = paste(c("reg_code", "reg_name"), rep(colnamecandidates[1], 2), sep = "_")

    fromstos = dplyr::as_tibble(froms)

  } else  if((length(headings)%% 2) == 0 & length(datastring) == 0) { # THen it is an empty update but we still need the "update dates"

    headingsinv = unlist(lapply(X= str_split(headings, " "), FUN = function(X){paste0(X[2]," ",  X[1], " 1")}))
    colnamecandidates = stringr::str_sub(
      stringr::str_replace_all(
        string =  lubridate::ymd(headingsinv),
        pattern = "-",
        replacement = ""),
      start=1, end = 6)

    froms = data.frame(
      matrix(
        data = c("a", "b"),
        ncol = 2, byrow = T)[NULL, ],
      stringsAsFactors = F)
    colnames(froms) =
      paste(c("reg_code", "reg_name"), rep(colnamecandidates[1], 2), sep = "_")

    tos = data.frame(
      matrix(data = c("a", "b"),
             ncol = 2, byrow = T)[NULL, ],
      stringsAsFactors = F)
    colnames(tos) = paste(c("reg_code", "reg_name"), rep(colnamecandidates[2], 2), sep = "_")

    fromstos = dplyr::bind_cols(dplyr::as_tibble(froms), dplyr::as_tibble(tos))
  } else {fromstos = NULL}


  return(fromstos)
}
regupdated = function(files){
  regiondef = region_at_time_txtfls(filename = files[1])

  for (i in 2:length(files)){
    print(i)
    regupdate = region_at_time_txtfls(filename = files[i])
    ## !! coming left_join: It would be best to find a way to join only by the "reg_code_x" variables but I could not find how to type this :-(
    both = dplyr::left_join(regiondef, regupdate)
    head(both)
    regupnames <- names(both)
    regupnamesl <- length(regupnames)
    both %>%
      dplyr::mutate(.,
                    !!sym(regupnames[regupnamesl-1]) :=
                      dplyr::case_when(
                        !is.na(!!dplyr::sym(regupnames[regupnamesl-1])) ~
                          !!dplyr::sym(regupnames[regupnamesl-1]),
                        TRUE ~ !!dplyr::sym(regupnames[regupnamesl-3])),
                    !!sym(regupnames[regupnamesl]) :=
                      dplyr::case_when(
                        !is.na(!!dplyr::sym(regupnames[regupnamesl])) ~ !!dplyr::sym(regupnames[regupnamesl]),
                        TRUE ~ !!dplyr::sym(regupnames[regupnamesl-2]))
      )  -> regiondef
  }
  return(regiondef)
}

no.regiontabell.flk = function(){
  #Fylker: https://www.ssb.no/en/klass/klassifikasjoner/104/versjon/1158/koder
  files <-   list.files( path = "./data-raw", pattern = ".txt", full.names = T) %>%
    .[which(!stringr::str_detect(., "~"))] %>%
    .[which(stringr::str_detect(., "Regindeling_Fylker"))]
  inndeling <- regupdated(files = files)
  return(inndeling)
}
#no.regiontabell.flk()

no.regiontabell.kmn = function(){
  #Kommuner: https://www.ssb.no/klass/klassifikasjoner/131
  files <-   list.files(path = "./data-raw", pattern = ".txt", full.names = T) %>%
    .[which(!stringr::str_detect(., "~"))] %>%
    .[which(stringr::str_detect(., "Regindeling_Kommuner"))]
  print(files)
  inndeling <- regupdated(files = files)
  return(inndeling)

}

regref_fylke <- no.regiontabell.flk()
regref_kommune <- no.regiontabell.kmn()

# creating a "long" mapping table between present region codes and current region codes.
prolong_region_mapping_table <- function(wide_region_mapping_table) {
  regref_codes <- wide_region_mapping_table %>% select(., starts_with("reg_code"))
  regref_names <- wide_region_mapping_table %>% select(., starts_with("reg_name"))

  start_states <- data.frame()
  for (i in sequence(ncol(regref_codes) )) {
    one_state_c = regref_codes[,i] %>%
      tidyr::pivot_longer(.,  cols = everything(), names_to = "time_from",  values_to = "region_code" ) %>%
      mutate(., ymfrom = stringr::str_extract(time_from, "\\d{6}"),
             yfrom = stringr::str_extract(ymfrom, "\\d{4}")) %>%
      select(., ymfrom, yfrom, reg_code_from = region_code)

    one_state_n = regref_names[,i] %>%
      tidyr::pivot_longer(.,  cols = everything(), names_to = "time_from",  values_to = "region_name" ) %>%
      mutate(., ymfrom = stringr::str_extract(time_from, "\\d{6}"),
             yfrom = stringr::str_extract(ymfrom, "\\d{4}")) %>%
      select(., reg_name_from = region_name)

    one_state <- dplyr::bind_cols(one_state_c, one_state_n)

    start_states <- dplyr::bind_rows(start_states, one_state)
  }

  end_states <- data.frame()
  for (i in sequence(ncol(regref_codes) )) {
    one_state_c = regref_codes[,ncol(regref_codes)] %>%
      tidyr::pivot_longer(.,  cols = everything(), names_to = "end_time",  values_to = "region_code" ) %>%
      mutate(., ymto = stringr::str_extract(end_time, "\\d{6}"),
             yto = stringr::str_extract(ymto, "\\d{4}")) %>%
      select(.,  ymto, yto, reg_code_to = region_code )

    one_state_n = regref_names[,ncol(regref_names)] %>%
      tidyr::pivot_longer(.,  cols = everything(), names_to = "end_year",  values_to = "region_name" ) %>%
      mutate(., ymto = stringr::str_extract(end_year, "\\d{6}"),
             yto = stringr::str_extract(ymto, "\\d{4}")) %>%
      select(.,  reg_name_to = region_name)

    one_state <- dplyr::bind_cols(one_state_c, one_state_n)
    end_states <- dplyr::bind_rows(end_states, one_state)
  }

  region_mapping_l <- dplyr::bind_cols(start_states, end_states)

  return(region_mapping_l)

}
regref_kommune_l = prolong_region_mapping_table(wide_region_mapping_table = regref_kommune)
regref_fylke_l = prolong_region_mapping_table(wide_region_mapping_table = regref_fylke)


## KPI
# https://www.ssb.no/statbank/table/03014
#metadt <- PxWebApiData::ApiData("http://data.ssb.no/api/v0/no/table/03014", returnMetaData = TRUE)
pxdt_t03014 <- PxWebApiData::ApiData("http://data.ssb.no/api/v0/no/table/03014",
                              Konsumgrp = "TOTAL",
                              Tid = T,
                              ContentsCode = "KpiAar")
kpi_t03014 <-
  dplyr::as.tbl(pxdt_t03014[[1]])  %>%
  dplyr::select(.,  -konsumgruppe, -statistikkvariabel) %>%
  dplyr::rename(., kpi = value)



m3_sortiment_kmn <-
  regnavn.at.ref.yr(
    regionstat = t03895(region_level = "kommune"),
    reg_level = "kommune", ref.yr = 2020 ) %>%
  dplyr::select(.,
                -tidyselect::starts_with("regref_"),
                -tidyselect::starts_with("regcode_"),
                -tidyselect::starts_with("yr_"),
                -tidyselect::starts_with("valid"))

m3_sortiment_flk <-
  regnavn.at.ref.yr2( regionstat = t03895(region_level = "fylke") ) %>%
  dplyr::select(., -regrefrow)

virkesverdi_kmn <-
  regnavn.at.ref.yr(
  regionstat = t03794( region_level = "kommune"),
  reg_level = "kommune") %>%
  dplyr::select(.,
                -tidyselect::starts_with("regref_"),
                -tidyselect::starts_with("regcode_"),
                -tidyselect::starts_with("yr_"),
                -tidyselect::starts_with("valid"))


virkesverdi_kmn2 <-
  regnavn.at.ref.yr2(
    regionstat = t03794( region_level = "kommune")) %>%
  dplyr::select(.,
                -regrefrow)


virkesverdi_flk <-
  regnavn.at.ref.yr2(
    regionstat = t03794( region_level = "fylke")) %>%
  dplyr::select(.,
                -regrefrow)


usethis::use_data(
  m3_sortiment_kmn,
  m3_sortiment_flk,
  virkesverdi_kmn,
  virkesverdi_flk,
  regref_kommune,
  regref_fylke,
  regref_kommune_l,
  regref_fylke_l,
  kpi_t03014,
  overwrite = T)
